"""Encoding and decoding ableton set xml tools."""

import textwrap
from typing import Literal, Tuple, Union, overload


class ParseError(Exception):
    """Error decoding/encoding."""


def xml_to_string(xml_str: str) -> Tuple[str, int]:
    """Strip xml hex data from ableton xml into single string."""
    if not xml_str:
        raise DecodeError(f"No text to parse! {xml_str}")
    levels = xml_str.splitlines()[1].count("\t")
    return xml_str.replace("\t", "").replace(" ", "").replace("\n", ""), levels


@overload
def hex_to_string(hex_str: str, return_bytes: Literal[False] = False) -> str:
    ...


@overload
def hex_to_string(hex_str: str, return_bytes: Literal[True]) -> bytes:
    ...


def hex_to_string(hex_str: str, return_bytes: bool = False) -> Union[str, bytes]:
    """Input is string, but it's actually encoded bytes in hex form.

    Decode hex to actual string representation.
    Format is character, null byte, character, null byte ...
    """
    byte_data = bytearray.fromhex(hex_str)
    if return_bytes:
        return byte_data
    return byte_data.decode("utf-16")


def string_to_hex(in_str: str) -> str:
    """Reverse of hex_to_string, take input string and turn into hex encoded string.

    Two extra Null bytes are at the end of the string(?)
    """
    byte_str = in_str.encode("utf-8")
    prepped = "".join([f"{x:x}00" for x in byte_str]).upper()
    return prepped + "0000"


def string_to_xml(in_str: str, levels: int = 14) -> str:
    """Reverse of xml_to_string, add in tabs and chunk data into 80 character wide block."""
    indent = "\t" * levels
    width = 80 + levels
    parsed = textwrap.wrap(
        in_str, initial_indent=indent, subsequent_indent=indent, break_long_words=True, tabsize=4, width=width
    )
    ending_indent = '\t' * (levels - 1)
    return "\n" + "\n".join(parsed) + f"\n{ending_indent}"



"""Define root level vars and functions."""

import pathlib
import sys
from typing import Dict, Literal, Optional, Tuple, Union, overload
from xml.etree import ElementTree as ET

import colorama

colorama.init(autoreset=True)

# These are the hex values of the drop down color menu, arranged in the same order of rows and columns.
# yapf: disable
# fmt: off
ableton_colors = [
    0xFF94A6, 0xFFA428, 0xCD9827, 0xF6F57C, 0xBEFA00, 0x21FF41, 0x25FEA9, 0x5DFFE9, 0x8AC5FE, 0x5480E4, 0x93A6FF, 0xD86CE4, 0xE552A1, 0xFFFEFE,
    0xFE3637, 0xF66D02, 0x99734A, 0xFEF134, 0x87FF67, 0x3DC201, 0x01BEAF, 0x18E9FE, 0x10A4EE, 0x007DC0, 0x886CE4, 0xB776C6, 0xFE38D4, 0xD1D0D1,
    0xE3665A, 0xFEA274, 0xD2AD70, 0xEDFFAE, 0xD3E499, 0xBAD175, 0x9AC58D, 0xD4FCE0, 0xCCF0F8, 0xB8C1E2, 0xCDBBE4, 0xAF98E4, 0xE5DDE0, 0xA9A8A8,
    0xC6938A, 0xB68257, 0x98826A, 0xBEBB69, 0xA6BE00, 0x7CB04C, 0x89C3BA, 0x9BB3C4, 0x84A5C3, 0x8392CD, 0xA494B5, 0xBF9FBE, 0xBD7096, 0x7B7A7A,
    0xAF3232, 0xA95131, 0x734E41, 0xDAC200, 0x84971F, 0x529E31, 0x0A9C8E, 0x236285, 0x1A2F96, 0x2E52A3, 0x624BAD, 0xA24AAD, 0xCD2E6F, 0xFFFEFE,
]


STEREO_OUTPUTS: Dict[int, Dict[str, str]] = {
    1: {"target": "AudioOut/External/S0",  "lower_display_string": "1/2"},
    2: {"target": "AudioOut/External/S1", "lower_display_string": "3/4"},
    3: {"target": "AudioOut/External/S2", "lower_display_string": "5/6"},
    4: {"target": "AudioOut/External/S3", "lower_display_string": "7/8"},
    5: {"target": "AudioOut/External/S4", "lower_display_string": "9/10"},
    6: {"target": "AudioOut/External/S5", "lower_display_string": "11/12"},
    7: {"target": "AudioOut/External/S6", "lower_display_string": "13/14"},
    8: {"target": "AudioOut/External/S7", "lower_display_string": "15/16"},
    9: {"target": "AudioOut/External/S8", "lower_display_string": "17/18"},
    10: {"target": "AudioOut/External/S9","lower_display_string": "19/20"},
}
# yapf: enable
# fmt: on


# Shorten color variables
RST = colorama.Fore.RESET
if sys.platform == "win32":
    BOLD = "[1m"  # pylint: disable=invalid-character-esc
    # Windows terminal colors are hard to see, use bright.
    R = colorama.Fore.LIGHTRED_EX
    G = colorama.Fore.LIGHTGREEN_EX
    B = colorama.Fore.LIGHTBLUE_EX
    Y = colorama.Fore.LIGHTYELLOW_EX
    C = colorama.Fore.LIGHTCYAN_EX
    M = colorama.Fore.LIGHTMAGENTA_EX
else:
    BOLD = "\033[1m"
    R = colorama.Fore.RED
    G = colorama.Fore.GREEN
    B = colorama.Fore.BLUE
    Y = colorama.Fore.YELLOW
    C = colorama.Fore.CYAN
    M = colorama.Fore.MAGENTA

RB = R + colorama.Style.BRIGHT + BOLD
GB = G + colorama.Style.BRIGHT + BOLD
BB = B + colorama.Style.BRIGHT + BOLD
YB = Y + colorama.Style.BRIGHT + BOLD
CB = C + colorama.Style.BRIGHT + BOLD
MB = M + colorama.Style.BRIGHT + BOLD

DEFAULT_DB_PATH = pathlib.Path.home() / "abletoolz_db.json"
BACKUP_DIR = "abletoolz_backup"


class AbletoolzError(Exception):
    """Error running Abletoolz."""


class ElementNotFound(Exception):
    """Element doesnt exist within the xml hierarchy where expected."""


@overload
def get_element(
    root: ET.Element,
    attribute_path: str,
    *,
    silent_error: Literal[False],
    attribute: Literal[None] = None,
) -> ET.Element:
    ...


@overload
def get_element(
    root: ET.Element,
    attribute_path: str,
    *,
    silent_error: Literal[True],
    attribute: Literal[None] = None,
) -> Optional[ET.Element]:
    ...


@overload
def get_element(
    root: ET.Element,
    attribute_path: str,
    *,
    silent_error: Literal[False] = False,
    attribute: str,
) -> str:
    ...


def get_element(
    root: ET.Element,
    attribute_path: str,
    *,
    silent_error: bool = False,
    attribute: Optional[str] = None,
) -> Union[ET.Element, str, None]:
    """Get element using Element tree xpath syntax."""
    element = root.findall(f"./{'/'.join(attribute_path.split('.'))}")
    if not element:
        if silent_error:
            return None
        # ElementTree.dump(root)
        raise ElementNotFound(f"{R}No element for path [{attribute_path}]")
    if attribute:
        attr = element[0].get(attribute)
        if attr is None:
            raise ElementNotFound(f"{R}Element {attribute}is empty!")
        return attr
    return element[0]


def note_translator(midi_note_number: int) -> Tuple[str, int]:
    """Return note and octave from midi note number."""
    notes = ["C", "C#", "D", "D#", "E", "F", "F#", "G", "G#", "A", "A#", "B"]
    octave = midi_note_number // 12 - 1
    note_index = midi_note_number % 12
    return notes[note_index], octave



"""Random util functions."""
import datetime
import logging
import os
import pathlib
from typing import Optional, Tuple
from xml.etree import ElementTree as ET

import pydantic

from abletoolz import decode_encode
from abletoolz.misc import BACKUP_DIR, B, ElementNotFound, R, Y, get_element

logger = logging.getLogger(__name__)


def format_date(timestamp: float) -> str:
    """Return easy to read date."""
    return datetime.datetime.fromtimestamp(timestamp).strftime("%m/%d/%Y %H:%M:%S")


def create_backup(pathlib_obj: pathlib.Path) -> None:
    """Move file to backup directory, does not replace previous files moved there."""
    backup_dir = pathlib_obj.parent / BACKUP_DIR
    backup_dir.mkdir(parents=True, exist_ok=True)
    ending_int = 1
    while True:
        backup_path = backup_dir / (pathlib_obj.stem + "__" + str(ending_int) + pathlib_obj.suffix)
        if pathlib.Path(backup_path).exists():
            ending_int += 1
        else:
            logger.info(
                "%sMoving original file to backup directory:\n%s --> %s",
                B,
                pathlib_obj,
                backup_path,
            )
            # Rename creates a new pathlib object with the new path,
            # so pathlib_obj will still point to the original path when the file is saved.
            pathlib_obj.rename(backup_path)
            return


def parse_mac_data(byte_data: bytes, abs_hash_path: str, debug: bool = False) -> Optional[str]:
    """Parse hex data for absolute path of file on MacOS.

    This hex byte data seems to be some type of struct with data_length_header:data:NULL sections of data, but other
    areas don't follow that pattern at all. This function attempts to grab all 'valid' data_length_header:data:NULL
    sections. The last two are always filepath, and the volume the file is on.
    """
    last_possible_index = len(byte_data) - 1
    itm_lst = []
    i = 6  # First 5 bytes are always a header containing the total amount of bytes.
    while i <= last_possible_index:
        if byte_data[i] == 0 or byte_data[i] == 0xFF:
            i += 1
            continue
        potential_length = byte_data[i]
        potential_end = i + potential_length + 1  # Potential ending NULL byte.
        if (
            potential_end < last_possible_index
            and 0 not in byte_data[i + 1 : potential_end]
            and byte_data[potential_end] == 0
        ):
            if debug:
                logger.info(
                    "\t%si: %s, value at i: %s, potential (e)ndex: %s, data: %s",
                    Y,
                    i,
                    byte_data[i],
                    potential_end,
                    byte_data[i + 1 : potential_end],
                )
            itm_lst.append(byte_data[i + 1 : potential_end])
            i = potential_end + 1
            continue
        i += 1
    try:
        return f"{itm_lst[-1].decode()}{itm_lst[-2].decode()}" if len(itm_lst) >= 2 else None
    except UnicodeDecodeError as e:
        logger.error("\n\n%s Couldn't decode: %s, Error: %s\n\n", R, abs_hash_path, e)
    return None


def parse_windows_data(byte_data: bytes, abs_hash_path: str) -> Optional[str]:
    r"""Parse hex byte data for absolute path of file on Windows.

    Windows hex bytes are utf-16 encoded with \x00 bytes in-between each character.
    """
    try:
        return byte_data.decode("utf-16").replace("\x00", "")  # Remove ending NULL byte.
    except UnicodeDecodeError as e:
        logger.error("\n\n%s Couldn't decode: %s, Error: %s\n\n", R, abs_hash_path, e)
    return None


def parse_hex_path(text: str) -> Optional[str]:
    """Take raw hex string from XML entry and parses."""
    if not text:
        return None
    # Strip new lines and tabs from raw text to have one long hex string.
    abs_hash_path = text.replace("\t", "").replace("\n", "")
    byte_data = bytearray.fromhex(abs_hash_path)
    if byte_data[0:3] == b"\x00" * 3:  # Header only on mac projects.
        # self.set_os = SetOperatingSystem.MAC_OS
        return parse_mac_data(byte_data, abs_hash_path)
    else:
        # self.set_os = SetOperatingSystem.WINDOWS_OS
        return parse_windows_data(byte_data, abs_hash_path)


def get_sample_size(file_ref: ET.Element) -> int:
    for file_size_str in ["OriginalFileSize", "FileSize"]:
        file_size = file_ref.findall(f".//{file_size_str}")
        # if file_size is None:
        #     raise ElementNotFound("Couldn't get sample size!")
        if len(file_size):
            try:
                return int(file_size[0].get("Value", ""))
            except ValueError as exc:
                raise ElementNotFound from exc
    logger.error("Couldn't find filesize!")
    return 0


def path_separator_type(path_str: str) -> str:
    """Get OS path string separator."""
    if "\\" in path_str:
        return "\\"
    elif "/" in path_str:
        return "/"
    else:
        raise Exception(f"Couldn't parse OS path type! {path_str}")


def check_relative_path(
    name: str,
    sample_element: ET.Element,
    project_root_folder: pathlib.Path,
) -> Tuple[Optional[pathlib.Path], Optional[pathlib.Path]]:
    """Constructs absolute path from project root and relative path stored in set."""
    if not project_root_folder:
        return None, None
    relative_path_enabled = get_element(sample_element, "FileRef.HasRelativePath", attribute="Value")
    relative_path_type = get_element(sample_element, "FileRef.RelativePathType", attribute="Value")
    if relative_path_enabled == "true" and relative_path_type == "3":
        relative_path_element = get_element(sample_element, "FileRef.RelativePath")
        sub_directory_path = []
        for path in relative_path_element:
            sub_directory_path.append(path.get("Dir"))
        from_project_root = f"{os.path.sep.join(sub_directory_path)}{os.path.sep}{name}"
        full_path = project_root_folder / os.path.sep.join(sub_directory_path) / name
        return full_path, from_project_root
    return None, None


class SampleRef(pydantic.BaseModel):
    """Sample ref model.

    Parse and recreate XML elements in these formats:

    11+ Example:
        <RelativePathType Value="3" />
        <RelativePath Value="Samples/Imported/2051 BASS.wav" />
        <Path Value="D:/AbletonSync/Drum N Bass/Samples/Imported/2051 BASS.wav" />
        <Type Value="1" />
        <LivePackName Value="" />
        <LivePackId Value="" />
        <OriginalFileSize Value="58394528" />
        <OriginalCrc Value="7866" />

    9 - 10 Example:
        <HasRelativePath Value="true" />
        <RelativePathType Value="3" />
        <RelativePath>
            <RelativePathElement Id="0" Dir="Samples" />
            <RelativePathElement Id="1" Dir="Processed" />
            <RelativePathElement Id="2" Dir="Freeze" />
        </RelativePath>
        <Name Value="Freeze 2-Drum Rack.wav" />
        <Type Value="1" />
        <Data>
            46003A005C004500780070006F00720074005C00410062006C00650074006F006E00200050007200
            6F006A0065006300740073005C0063006800610072006700650064002000500072006F006A006500
            630074005C00530061006D0070006C00650073005C00500072006F00630065007300730065006400
            5C0046007200650065007A0065005C0046007200650065007A006500200032002D00440072007500
            6D0020005200610063006B002E007700610076000000
        </Data>
        <RefersToFolder Value="false" />
        <SearchHint>
            <PathHint>
            </PathHint>
            <FileSize Value="0" />
            <Crc Value="0" />
            <MaxCrcSize Value="0" />
            <HasExtendedInfo Value="false" />
        </SearchHint>
    """

    name: str
    size: int
    last_modified: int
    crc: int
    relative_type_element: ET.Element
    sample_ref: ET.Element
    absolute_element: ET.Element
    relative_element: ET.Element
    version_tuple: Tuple[int, int, int]

    absolute: Optional[pathlib.Path] = None
    relative: Optional[pathlib.Path] = None
    project_root: Optional[pathlib.Path] = None

    class Config:
        """Pydantic config."""

        arbitrary_types_allowed = True

    @classmethod
    def from_element(
        cls,
        sample_ref: ET.Element,
        version_tuple: Tuple[int, int, int],
        project_root_folder: pathlib.Path,
    ) -> "SampleRef":
        """Parse ElementTree into class.

        11 > switched to simple string paths, before 11 used binary encoded data, which
        differs for Windows and MacOs.
        """
        last_modified = sample_ref.find("LastModDate").get("Value")
        file_ref = sample_ref.find("FileRef")
        file_size = get_sample_size(file_ref)
        relative_type_element = file_ref.find("RelativePathType")
        if version_tuple >= (11, 0, 0):
            absolute_element = file_ref.find("Path")
            crc = file_ref.find("OriginalCrc").get("Value")
            absolute = file_ref.find("Path").get("Value")
            relative_element = file_ref.find("RelativePath")
            relative = file_ref.find("RelativePath").get("Value")
            name = pathlib.Path(absolute).name
        else:
            absolute_element = file_ref.find("Data")
            absolute = parse_hex_path(absolute_element.text)
            name = file_ref.find("Name").get("Value", "")
            try:
                crc = file_ref.findall(".//Crc")[0].get("Value")
            except IndexError:
                crc = 0
                ET.dump(file_ref)
            relative, _ = check_relative_path(name, sample_ref, project_root_folder)
            relative_element = file_ref.find("RelativePath")

        return cls(
            name=name,
            size=file_size,
            last_modified=last_modified,
            relative_type_element=relative_type_element,
            relative_element=relative_element,
            sample_ref=sample_ref,
            crc=crc,
            absolute_element=absolute_element,
            absolute=pathlib.Path(absolute) if absolute else None,
            relative=pathlib.Path(relative) if relative else None,
            project_root=project_root_folder,
            version_tuple=version_tuple,
        )

    @property
    def absolute_exists(self) -> bool:
        return self.absolute and self.absolute.exists()

    @property
    def relative_exists(self) -> bool:
        return self.relative and self.project_root and (self.project_root / self.relative).exists()

    def get_original_file_ref(self) -> ET.Element:
        return get_element(self.sample_ref, "SourceContext.SourceContext.OriginalFileRef.FileRef")

    def set_absolute(self, path: pathlib.Path) -> None:
        """Set absolute path for sample ref xml element."""
        if self.version_tuple < (11, 0, 0):
            # Get indentation level from current xml data.
            _, levels = decode_encode.xml_to_string(self.absolute_element.text)
            # Set new value.
            hex_string = decode_encode.string_to_hex(str(path))
            formatted_xml = decode_encode.string_to_xml(hex_string, levels=levels)
            self.absolute_element.text = formatted_xml
            try:
                second_ref = get_element(self.sample_ref, "SourceContext.SourceContext.OriginalFileRef.FileRef.Data")
            except ElementNotFound:
                return
            _, levels = decode_encode.xml_to_string(second_ref.text)
            formatted_xml = decode_encode.string_to_xml(hex_string, levels=levels)
            second_ref.text = formatted_xml

            return

        self.absolute_element.set("Value", str(path))

    def set_relative(self, path: str) -> None:
        """Set relative path from project root.

        Pre 11
            <HasRelativePath Value="true" />
            <RelativePath>
                <RelativePathElement Dir="Samples" />
                <RelativePathElement Dir="Imported" />
            </RelativePath>

        Post 11
            <RelativePathType Value="3" />
            <RelativePath Value="Samples/Imported/2051 BASS.wav" />
        """
        if self.version_tuple < (11, 0, 0):
            # Clear out old entries
            old = [e for e in self.relative_element]
            tails = [x.tail for x in old]
            for e in old:
                self.relative_element.remove(e)
            for i, folder in enumerate(path.split("/")[:-1]):
                element = ET.Element("RelativePathElement", attrib=dict(Id=i, Dir=folder))
                element.tail = tails[i]
                self.relative_element.append(element)
            return
        self.relative_element.set("Value", path)

    def get_relative_value(self) -> pathlib.Path:
        """Get path from relative xml element."""
        if self.version_tuple < (11, 0, 0):
            sub_directory_path = []
            for path in self.relative_element:
                sub_directory_path.append(path.get("Dir"))
            return pathlib.Path("/".join(sub_directory_path))
        return pathlib.Path(self.relative_element.get("Value")).parent

    def set_relative_type(self, type_int: int) -> None:
        """Set relative path type.

        <RelativePathType Value="3" />
        1 or 0 is absolute (?)
        3 is relative from project root
        """
        if self.version_tuple < (11, 0, 0):
            try:
                has_rel_ele = get_element(self.sample_ref, "FileRef.HasRelativePath")
                if type_int == 3:
                    has_rel_ele.set("Value", "true")
                elif type_int in {0, 1}:
                    has_rel_ele.set("Value", "false")
            except ElementNotFound:
                pass
        self.relative_type_element.set("Value", str(type_int))

    def get_relative_type(self) -> int:
        """Get relative path type (integer)."""
        return int(self.relative_type_element.get("Value"))

    def clear_search_hints(self) -> None:
        """Remove search hints, which are the sample paths to folders in abletons browser."""
        # search_hints = self.sample_ref.findall("SearchHint")
        for search_hint in self.sample_ref.iter("SearchHint"):
            refs = [e for e in search_hint]
            for ref in refs:
                search_hint.remove(ref)



"""Ableton track parser."""
import logging
from typing import Iterator, Tuple
from xml.etree import ElementTree as ET

from abletoolz.misc import B, C, G, M, get_element

logger = logging.getLogger(__name__)


class AbletonTrack(object):
    """Single track object."""

    def __init__(self, track_root: ET.Element, version: Tuple[int, int, int]) -> None:
        """Construct AbletonTrack."""
        self.track_root = track_root
        self.type = track_root.tag
        self.name = get_element(track_root, "Name.UserName", attribute="Value")
        if not self.name:
            self.name = get_element(track_root, "Name.EffectiveName", attribute="Value")
        self.id = track_root.get("Id")
        self.group_id = get_element(track_root, "TrackGroupId", attribute="Value")
        # Guessing 'Sesstion' was a typo early on and got stuck to not break backwards compatibility
        self.width = get_element(
            track_root,
            "DeviceChain.Mixer.ViewStateSesstionTrackWidth",
            attribute="Value",
        )
        # Lane height in arrangement view will be automation lane 0
        self.height = get_element(
            track_root,
            "DeviceChain.AutomationLanes.AutomationLanes.AutomationLane.LaneHeight",
            attribute="Value",
        )
        # Ableton 11 changes.
        self.color_element = "Color" if version > (11, 0, 0) else "ColorIndex"
        self.unfolded = get_element(track_root, "TrackUnfolded", attribute="Value", silent_error=True)  # Ableton 10
        if not self.unfolded:
            folded = get_element(track_root, "DeviceChain.Mixer.IsFolded", attribute="Value")  # Ableton 9/8
            self.unfolded = "false" if folded == "true" else "true"

    def __str__(self) -> str:
        """Create string representation of parsed track."""
        return (
            f"{B}Track type {self.type:>12}, {G}Name {self.name:>50}, {C}Id {self.id:>4}, "
            f"Group id {self.group_id:>4}, {M}Color {self.color:>3}, Width {self.width:>3}, "
            f"Height {self.height:>3}, Unfolded: {self.unfolded}"
        )

    @property
    def color(self) -> int:
        """Return color.

        There are 70 possible color values, color menu has 5 rows of 14

        Returns:
            -1 on failure to get element value, color index otherwise.
        """
        if (clr := self.track_root.find(self.color_element)) is not None:
            return int(clr.get("Value", 0))
        return -1

    @color.setter
    def color(self, value: int) -> None:
        """Set color for track."""
        if not 0 <= value <= 69:
            raise ValueError("Color index must be within 0 - 69")
        if (clr_element := self.track_root.find(self.color_element)) is not None:
            clr_element.set("Value", str(value))

    def clips_clipview(self) -> Iterator[ET.Element]:
        """Iterate through all the clips in the clip view."""
        for clip in self.track_root.iter("ClipSlot"):
            yield clip

    def clips_arrangement(self) -> Iterator[ET.Element]:
        """Iterate through all the clips in the arangement view."""
        if self.type == "MidiTrack":
            tree_element = "ClipTimeable"
            clip_element = "MidiClip"
        elif self.type == "AudioTrack":
            tree_element = "Sample"
            clip_element = "AudioClip"
        else:
            return
        clip_timeample = self.track_root.find(f".//{tree_element}")
        if not clip_timeample:
            return
        for arrange_clip in clip_timeample.iter(clip_element):
            yield arrange_clip

    def clip_clipview_colors(self) -> Iterator[ET.Element]:
        """Yield clipview color elements from current track."""
        for clip in self.clips_clipview():
            if (color_element := clip.find(f".//{self.color_element}")) is not None:
                yield color_element

    def clip_arangement_colors(self) -> Iterator[ET.Element]:
        """Yield arrangement color elements from current track."""
        for clip in self.clips_arrangement():
            if (color_element := clip.find(f".//{self.color_element}")) is not None:
                yield color_element



import logging
import pathlib
import sys
import xml.etree.ElementTree as ET
from typing import Optional, Tuple, Union

from pydantic import BaseModel

sys.path.append(r"C:\Users\elixer\repos\_abletoolz")
# sys.path.append("/Users/fivehtp/repos/_abletoolz")
from abletoolz.ableton_set import AbletonSet
from abletoolz import decode_encode

logger = logging.getLogger(__name__)


def parse_vst_element(vst_element: ET.Element) -> Tuple[Optional[pathlib.Path], Optional[str], Optional[pathlib.Path]]:
    def handle_path_element(path_result):
        full_path = path_result.get("Value")
        if not full_path:
            logger.error("Couldn't get Path for %s", path_result)
            return None, None, None

        if not ("/" in full_path or "\\" in full_path):
            search_result = self.search_plugins(full_path)
            if search_result:
                return None, search_result.name, search_result
            return None, full_path, None

        path_separator = self.path_separator_type(full_path)
        name = full_path.split(path_separator)[-1]
        return pathlib.Path(full_path), name, None

    def handle_dir_element(path_result):
        dir_bin = path_result.find("Data")
        if not dir_bin:
            logger.error("Couldn't get Path for %s", path_result)
            return None, None, None

        text = dir_bin.text
        if not text:
            return None, None, None

        path = self._parse_hex_path(text)
        name_ele = vst_element.find("FileName")
        name = name_ele.get("Value", "") if name_ele is not None else "<>"

        if not path:
            logger.error("%sCouldn't parse absolute path for %s", Y, name)
            return None, name, None

        path_separator = self.path_separator_type(path)
        full_path = f"{path}{path_separator}{name}" if path[-1] != path_separator else f"{path}{name}"

        return pathlib.Path(full_path), name, None

    for plugin_path in ["Dir", "Path"]:
        path_results = vst_element.findall(f".//{plugin_path}")
        if not path_results:
            continue

        if plugin_path == "Path":
            return handle_path_element(path_results[0])
        elif plugin_path == "Dir":
            return handle_dir_element(path_results[0])

    logger.error("%sCouldn't parse plugin!", R)
    return None, None, None


class PluginData(BaseModel):
    data: Optional[str]
    file_name: Optional[str]
    plug_name: Optional[str]
    unique_id: Optional[int]
    buffer: Optional[str]


def parse_plugin_data(root: ET.Element) -> PluginData:
    """."""
    dir_element = root.find("Dir")
    path_element = root.find("Path")

    path = path_element.get("Value") if path_element is not None else None
    data = dir_element.findtext("Data") if dir_element is not None else None
    # decode_encode.

    file_name = root.find("FileName").get("Value") if root.find("FileName") is not None else None
    plug_name = root.find("PlugName").get("Value") if root.find("PlugName") is not None else None

    unique_id_str = root.findtext("UniqueId")
    unique_id = int(unique_id_str) if unique_id_str else None

    vst_preset = root.find(".//VstPreset")
    buffer = vst_preset.findtext("Buffer") if vst_preset is not None else None

    return PluginData(data=data, file_name=file_name, plug_name=plug_name, unique_id=unique_id, buffer=buffer)





als_set = pathlib.Path(
    "../test/sample_missing_fix Project/11.2.10_11 bass_pad fm8_350bars_172.00bpm.als"
)

ableton_set = AbletonSet(als_set)
ableton_set.parse()
plugins = []
for plugin_element in ableton_set.root.iter("PluginDesc"):
    for vst_element in plugin_element.iter("VstPluginInfo"):
        plugins.append(parse_plugin_data(vst_element))

print(plugins)



"""Tools for dealing with ableton colors, creating gradients etc."""
import logging
import random
from typing import Final, List, Optional, Tuple

import numpy as np
from colormath.color_conversions import convert_color
from colormath.color_objects import LabColor, sRGBColor

logger = logging.getLogger(__name__)
# These are the hex values of the drop down color menu, matching the exact layout, hence the lines being too long.
# yapf: disable
ableton_colors_strs: Final = [
    "#FF94A6", "#FFA428", "#CD9827", "#F6F57C", "#BEFA00", "#21FF41", "#25FEA9", "#5DFFE9", "#8AC5FE", "#5480E4", "#93A6FF", "#D86CE4", "#E552A1", "#FFFEFE",
    "#FE3637", "#F66D02", "#99734A", "#FEF134", "#87FF67", "#3DC201", "#01BEAF", "#18E9FE", "#10A4EE", "#007DC0", "#886CE4", "#B776C6", "#FE38D4", "#D1D0D1",
    "#E3665A", "#FEA274", "#D2AD70", "#EDFFAE", "#D3E499", "#BAD175", "#9AC58D", "#D4FCE0", "#CCF0F8", "#B8C1E2", "#CDBBE4", "#AF98E4", "#E5DDE0", "#A9A8A8",
    "#C6938A", "#B68257", "#98826A", "#BEBB69", "#A6BE00", "#7CB04C", "#89C3BA", "#9BB3C4", "#84A5C3", "#8392CD", "#A494B5", "#BF9FBE", "#BD7096", "#7B7A7A",
    "#AF3232", "#A95131", "#734E41", "#DAC200", "#84971F", "#529E31", "#0A9C8E", "#236285", "#1A2F96", "#2E52A3", "#624BAD", "#A24AAD", "#CD2E6F", "#FFFEFE",
]
# yapf: enable
ableton_colors = [int(x.strip("#"), 16) for x in ableton_colors_strs]
colors_indexed = [(i, x) for i, x in enumerate(ableton_colors)]
# TODO: Build up this list with good results to re-use.
known_good_combos = {
    "light_blue_to_dark_pink": (0x5DFFE9, 0xE552A1),
    "pink_to_yellow": (0xD86CE4, 0xFEF134),
    "yellow_green_to_dark_green": (0xBEFA00, 0x3DC201),
}


def hex_to_rgb(hex_value: int) -> Tuple[int, int, int]:
    """Convert hex to rgb."""
    return ((hex_value >> 16) & 0xFF, (hex_value >> 8) & 0xFF, hex_value & 0xFF)


def rgb_to_hex(r: int, g: int, b: int) -> int:
    """Convert rgb values to hex integer."""
    return (r << 16) + (g << 8) + b


def interpolate_color(color1: int, color2: int, t: float) -> int:
    """Find color inbetween two given."""
    r1, g1, b1 = hex_to_rgb(color1)
    r2, g2, b2 = hex_to_rgb(color2)
    r = int(r1 * (1 - t) + r2 * t)
    g = int(g1 * (1 - t) + g2 * t)
    b = int(b1 * (1 - t) + b2 * t)
    return rgb_to_hex(r, g, b)


def create_gradient(color1: int, color2: int, steps: int) -> List[int]:
    """Generate a color gradient between two colors."""
    return [interpolate_color(color1, color2, t / (steps - 1)) for t in range(steps)]


def hex_to_lab_color(hex_value: int) -> LabColor:
    """Convert hex value to colormath lab color."""
    r, g, b = hex_to_rgb(hex_value)
    rgb_color = sRGBColor(r, g, b, is_upscaled=True)
    lab_color = convert_color(rgb_color, LabColor)
    return lab_color


def custom_delta_e_cie2000(color1: LabColor, color2: LabColor) -> float:
    """Re-implementation of colormath function that is currently broken due to using deprecated numpy methods."""
    c_1 = np.sqrt(color1.lab_a**2 + color1.lab_b**2)
    c_2 = np.sqrt(color2.lab_a**2 + color2.lab_b**2)

    c_mean = (c_1 + c_2) / 2
    g = 0.5 * (1 - np.sqrt(c_mean**7 / (c_mean**7 + 25**7)))

    a1_prime = (1 + g) * color1.lab_a
    a2_prime = (1 + g) * color2.lab_a

    c1_prime = np.sqrt(a1_prime**2 + color1.lab_b**2)
    c2_prime = np.sqrt(a2_prime**2 + color2.lab_b**2)

    h1_prime = np.rad2deg(np.arctan2(color1.lab_b, a1_prime))
    h2_prime = np.rad2deg(np.arctan2(color2.lab_b, a2_prime))
    h1_prime += 360 * (h1_prime < 0)
    h2_prime += 360 * (h2_prime < 0)

    delta_l = color2.lab_l - color1.lab_l
    delta_c = c2_prime - c1_prime
    delta_h = h2_prime - h1_prime

    delta_h -= 360 * (delta_h > 180)
    delta_h += 360 * (delta_h < -180)

    delta_h_bar = np.abs(h1_prime - h2_prime)
    delta_h_bar -= 360 * (delta_h_bar > 180)
    delta_h_bar += 360 * (delta_h_bar < -180)

    h_bar = (h1_prime + h2_prime) / 2
    h_bar += 360 * (h_bar < 0)
    h_bar -= 360 * (h_bar > 180)

    t = (
        1
        - 0.17 * np.cos(np.deg2rad(h_bar - 30))
        + 0.24 * np.cos(np.deg2rad(2 * h_bar))
        + 0.32 * np.cos(np.deg2rad(3 * h_bar + 6))
        - 0.20 * np.cos(np.deg2rad(4 * h_bar - 63))
    )

    delta_h_prime = 2 * np.sqrt(c1_prime * c2_prime) * np.sin(np.deg2rad(delta_h_bar) / 2)
    sl = 1 + (0.015 * (color1.lab_l - 50) ** 2) / np.sqrt(20 + (color1.lab_l - 50) ** 2)
    sc = 1 + 0.045 * c_mean
    sh = 1 + 0.015 * c_mean * t
    rt = (
        -2
        * np.sqrt(c_mean**7 / (c_mean**7 + 25**7))
        * np.sin(np.deg2rad(60 * np.exp(-(((h_bar - 275) / 25) ** 2))))
    )

    delta_e = np.sqrt(
        (delta_l / sl) ** 2
        + (delta_c / sc) ** 2
        + (delta_h_prime / sh) ** 2
        + rt * (delta_c / sc) * (delta_h_prime / sh)
    )
    return delta_e


def find_closest_color_ciede2000(color: int, palette: List[Tuple[int, int]]) -> Tuple[int, int]:
    """Find closest color using ciede2000."""
    lab_color = hex_to_lab_color(color)
    closest_color = min(palette, key=lambda x: custom_delta_e_cie2000(lab_color, hex_to_lab_color(x[1])))
    return closest_color


def find_furthest_color_ciede2000(color: int, palette: List[Tuple[int, int]]) -> Tuple[int, int]:
    """Find furthest color using ciede2000."""
    lab_color = hex_to_lab_color(color)
    furthest_color = max(palette, key=lambda x: custom_delta_e_cie2000(lab_color, hex_to_lab_color(x[1])))
    return furthest_color


def create_gradient_from_palette_ciede2000(
    color1: int, color2: int, steps: int, palette: List[int]
) -> List[Tuple[int, int]]:
    """Generate a gradient between two colors, and use ciede2000 to find the closest available from a palette."""
    gradient = create_gradient(color1, color2, steps)
    indexed_palette = [(index, color) for index, color in enumerate(palette)]
    return [find_closest_color_ciede2000(color, indexed_palette) for color in gradient]


def sort_colors_ciede2000(colors: List[int], start_color_index: int) -> List[Tuple[int, int]]:
    """Sort colors based on the ciede2000 algorithm which is more accurate to human perception."""
    start_color = colors[start_color_index]
    sorted_colors = [(start_color_index, start_color)]
    remaining_colors = [(index, color) for index, color in enumerate(colors) if index != start_color_index]

    while remaining_colors:
        last_color = sorted_colors[-1][1]
        closest_color_index, closest_color = find_closest_color_ciede2000(last_color, remaining_colors)
        sorted_colors.append((closest_color_index, closest_color))
        remaining_colors.remove((closest_color_index, closest_color))

    return sorted_colors


def create_gradient_ableton(
    num_items: int, starting_color: Optional[int] = None, starting_index: Optional[int] = None
) -> List[int]:
    """Makes gradients for ableton tracks/clips.

    Since there are only 70 colors, I try to find the furthest color from the first one, which can be
    random or selected. If the number of items is too short, this will be a large jump so set min to 5.

    Returns the ableton color index, which is the index of the array defined at the top of this file. It maps to
    the colors in the ableton menu.
    """
    num_items = max(10, num_items)
    if starting_index is not None:
        color_1 = ableton_colors[starting_index]
    else:
        color_1 = starting_color if starting_color is not None else random.sample(ableton_colors, 1)[0]

    color_2 = find_furthest_color_ciede2000(color_1, colors_indexed)[1]
    natural_grad = create_gradient_from_palette_ciede2000(color_1, color_2, num_items, ableton_colors)
    return [c[0] for c in natural_grad]



"""Ableton set parsing."""
import enum
import functools
import gzip
import logging
import os
import pathlib
import re
import shutil
import subprocess
import sys
import threading
from typing import Callable, Dict, Generator, List, Optional, ParamSpec, Tuple, TypeVar
from xml.etree import ElementTree as ET

from abletoolz import color_tools, utils
from abletoolz.ableton_track import AbletonTrack
from abletoolz.misc import CB, RB, RST, STEREO_OUTPUTS, B, C, G, M, R, Y, get_element

if sys.platform == "win32":
    import win32_setctime

logger = logging.getLogger(__name__)


class SetError(Exception):
    """Ableton set parse error."""


class SetOperatingSystem(enum.Enum):
    """Pre ableton 11, sets store data differently.

    Sets do not store any OS information, but we can guess based on encoding of data, AU units vs VSTs and some
    other differences.
    """

    MAC_OS = enum.auto()
    WINDOWS_OS = enum.auto()
    UNSET = enum.auto()


def version_supported(set_version: Tuple[int, int, int], supported_version: Tuple[int, int, int]) -> bool:
    """Check if set version is supported for method."""
    for set_v, supported_v in zip(set_version, supported_version):
        if set_v > supported_v:
            return True
        elif set_v < supported_v:
            return False
    return True


P = ParamSpec("P")
RT = TypeVar("RT")


def above_version(supported_version: Tuple[int, int, int]) -> Callable[[Callable[P, RT]], Callable[P, RT]]:
    """Decorator factory to handle method support for changing XML schemas across Ableton versions.

    https://help.ableton.com/hc/en-us/articles/360000841004-Backward-Compatibility
    """

    def wrapper(f: Callable[P, RT]) -> Callable[P, RT]:
        @functools.wraps(f)
        def wrapped_func(*args: P.args, **kwargs: P.kwargs) -> RT:
            # arg[0] is self
            if not version_supported(args[0].version_tuple, supported_version):  # type: ignore[attr-defined]
                logger.info("Function %s is only supported for %s and above.", f.__name__, supported_version)
                raise SetError(f"{f.__name__} not supported for this version!")
            return f(*args, **kwargs)

        return wrapped_func

    return wrapper


def set_loaded(f: Callable[..., RT]) -> Callable[..., RT]:
    """Decorator that checks set XML has been loaded into Xtree."""

    @functools.wraps(f)
    def wrapped_func(self: "AbletonSet", *args, **kwargs) -> RT:
        if self.root is None:
            raise SetError("Set is not loaded!")
        return f(self, *args, **kwargs)

    return wrapped_func


def elements_equal(e1: ET.Element, e2: ET.Element) -> bool:
    """Check if two xml.Etree roots are equivalent."""
    if e1.tag != e2.tag:
        return False
    if e1.text != e2.text:
        return False
    if e1.tail != e2.tail:
        return False
    if e1.attrib != e2.attrib:
        return False
    if len(e1) != len(e2):
        return False
    return all(elements_equal(c1, c2) for c1, c2 in zip(e1, e2))


class AbletonSet(object):
    """Set object."""

    def __init__(self, pathlib_obj: pathlib.Path) -> None:
        """Construct class."""
        self.name = pathlib_obj.name
        self.path = pathlib_obj
        self.tree = None
        self.root: Optional[ET.Element] = None

        # Parsed set variables.
        self.project_root_folder: Optional[pathlib.Path] = None  # Folder where Ableton Project Info resides.
        self.last_modification_time: Optional[float] = None
        self.creation_time: Optional[float] = None

        self.version: Optional[str] = None  # Official Ableton live version.
        self.version_tuple: Optional[Tuple[int, int, int]] = None

        self.tempo: Optional[float] = None
        self.furthest_bar: Optional[int] = None
        self.bpm: Optional[float] = None
        self.tracks: List[AbletonTrack] = []
        self.sample_list: List[utils.SampleRef] = []

        # TODO WIP, not finished/used necessarily.
        self.set_os: SetOperatingSystem = SetOperatingSystem.UNSET
        self.missing_absolute_samples: List[pathlib.Path] = []
        self.missing_relative_samples: List[pathlib.Path] = []
        self.found_vst_dirs: List[pathlib.Path] = []
        self.last_elem = None
        self.key = None

    def __eq__(self, o: object) -> bool:
        """Compare two sets."""
        if self.root is None:
            return False
        if not isinstance(o, AbletonSet) or isinstance(o, AbletonSet) and o.root is None:
            return False
        return elements_equal(self.root, o.root)

    def open_folder(self) -> None:
        """Open folder in file explorer/finder.

        Currently unused.
        """
        if sys.platform == "win32":
            subprocess.Popen(f'explorer /select, "{self.path}"')
        elif sys.platform == "darwin":
            subprocess.Popen(f"open {self.path}")

    @set_loaded
    def load_version(self) -> None:
        """Load version."""
        assert self.root  # Shut mypy up since decorator checks this.
        self.version = self.root.get("Creator")
        if not isinstance(self.version, str):
            raise SetError("Couldn't parse Creator from set.")
        parsed = re.findall(
            r"Ableton Live ([0-9]{1,2})\.([0-9]{1,3})[\.b]{0,1}([0-9]{1,3}){0,1}",
            self.version,
        )
        if not parsed:
            raise SetError("Couldn't parse set version!")
        parsed = [int(x) if x.isdigit() else x for x in parsed[0] if x != ""]
        if len(parsed) == 3:
            major, minor, patch = parsed
        elif len(parsed) == 2:
            major, minor = parsed
            patch = 0
        else:
            raise SetError(f"Could not parse version from: {self.version}")
        self.version_tuple = major, minor, patch
        logger.info("%sSet version: %s%s", B, M, self.version)
        if "b" in self.version.split()[-1]:
            logger.warning("%sSet is from a beta version, some commands might not work properly!", Y)

    def parse(self) -> bool:
        """Uncompresses ableton set and loads into element tree."""
        with open(self.path, "rb") as fd:
            first_two_bytes = fd.read(2)
            if first_two_bytes == b"\xab\x1e":  # yes, it spells able :P
                logger.error("%s%sIs pre Ableton 8.2.x which is unsupported.", R, self.path)
                return False
            elif first_two_bytes != b"\x1f\x8b":
                logger.error(
                    "%s%sFile is not .als or is an older format that doesn't use gzip!, cannot open...", R, self.path
                )
                return False
        self.get_file_times()
        with gzip.open(self.path, "r") as fd:
            data = fd.read().decode("utf-8")
            if not data:
                logger.error("%sError loading data %s!", R, self.path)
                return False
            self.root = ET.fromstring(data)
            return True

    def find_project_root_folder(self) -> Optional[pathlib.Path]:
        """Find project root folder for set."""
        # TODO Parse project .cfg file and logger.info information.
        if self.project_root_folder:
            return self.project_root_folder

        max_folder_search_depth = 10
        for i, current_dir in enumerate(self.path.parents):
            if i > max_folder_search_depth:
                logger.warning("%sReached maximum search depth, exiting..", R)
                break
            elif pathlib.Path(current_dir / "Ableton Project Info").exists():
                self.project_root_folder = current_dir
                logger.debug("%sProject root folder: %s", C, current_dir)
                return self.project_root_folder
        logger.error("%sCould not find project folder(Ableton Project Info), unable to validate relative paths!", R)
        return None

    def generate_xml(self) -> bytes:
        """Add header and footer to xml data."""
        if self.root is None:
            raise SetError("Set is not loaded!")
        header = '<?xml version="1.0" encoding="UTF-8"?>\n'.encode("utf-8")
        footer = "\n".encode("utf-8")
        xml_output = ET.tostring(self.root, encoding="utf-8")
        return header + xml_output + footer

    def save_xml(self) -> None:
        """Save set XML."""
        xml_file = self.path.parent / (self.path.stem + ".xml")
        if xml_file.exists():
            utils.create_backup(xml_file)
        with xml_file as fd:
            fd.write_bytes(self.generate_xml())
        logger.info("%sSaved xml to %s", G, xml_file)

    def get_file_times(self) -> None:
        """Find set creation/modification times."""
        if sys.platform == "win32":
            self.creation_time = os.path.getctime(self.path)
        else:
            self.creation_time = os.stat(self.path).st_birthtime
        self.last_modification_time = os.path.getmtime(self.path)
        logger.debug(
            "%sFile creation time %s, Last modification time: %s",
            B,
            utils.format_date(self.creation_time),
            utils.format_date(self.last_modification_time),
        )

    def restore_file_times(self, pathlib_obj: pathlib.Path) -> None:
        """Restore original creation and modification times to file."""
        if self.last_modification_time is None:
            logger.warning("No modification time! Can't restore original time...")
            return
        os.utime(self.path, (self.last_modification_time, self.last_modification_time))
        if sys.platform == "win32":
            win32_setctime.setctime(self.path, self.creation_time)
        elif sys.platform == "darwin":
            date = utils.format_date(self.creation_time)
            path = str(pathlib_obj).replace(" ", r"\ ")
            os.system(f'SetFile -d "{date}" {path} >/dev/null')
        logger.debug(
            "%sRestored set creation and modification times: %s, %s",
            G,
            utils.format_date(self.creation_time),
            utils.format_date(self.last_modification_time),
        )

    def write_set(self) -> None:
        """Recompresses set to gzip. Used in thread to help prevent file getting corrupted mid write."""
        with gzip.open(self.path, "wb") as fd:
            fd.write(self.generate_xml())
        logger.info("%sSaved set to %s", G, self.path)
        self.restore_file_times(self.path)

    def save_set(self, append_bars_bpm: bool = False, prepend_version: bool = False) -> None:
        """Save set to disk with optional filename modifications.

        This function saves the current set to disk, first creating a backup of the original file.
        It optionally appends the number of bars and BPM to the filename, and/or prepends the version number.
        The actual writing of the set to disk is performed in a separate thread.

        Args:
            append_bars_bpm: If True, append the number of bars and BPM to the filename.
            prepend_version: If True, prepend the version number to the filename.
        """
        utils.create_backup(self.path)
        if append_bars_bpm:
            cleaned_name = re.sub(r"_\d{1,3}bars_\d{1,3}\.\d{2}bpm", "", self.path.stem)
            new_filename = cleaned_name + f"_{self.furthest_bar}bars_{self.bpm:.2f}bpm.als"
            self.path = pathlib.Path(self.path.parent / new_filename)
            logger.debug("%sAppending bars and bpm, new set name: %s.als", M, self.path.stem)

        if self.version_tuple and prepend_version:
            version_string = f"{self.version_tuple[0]}.{self.version_tuple[1]}.{self.version_tuple[2]}_"
            cleaned_name = re.sub(r"\d{1,2}\.\d{1,3}\.[b\d]{1,5}_", "", self.path.stem)
            self.path = self.path.parent / (version_string + cleaned_name + self.path.suffix)

        # Create non daemon thread so that it is not forcibly killed if parent process is killed.
        thread = threading.Thread(target=self.write_set)
        thread.start()
        thread.join()

    # Data parsing functions.
    @above_version(supported_version=(8, 0, 0))
    def load_tracks(self) -> None:
        """Load tracks into AbletonTrack src."""
        tracks = get_element(self.root, "LiveSet.Tracks")
        for track in tracks:
            self.tracks.append(AbletonTrack(track, self.version_tuple))

    def print_tracks(self) -> None:
        """logger.infos track info."""
        logger.info("Tracks:\n%s", "\n".join([str(x) for x in self.tracks]))

    @set_loaded
    def find_furthest_bar(self) -> int:
        """Find the max of the longest clip or furthest bar something is in Arrangement."""
        assert self.root is not None  # Shut mypy up.
        current_end_times = [int(float(end_times.get("Value", 0))) for end_times in self.root.iter("CurrentEnd")]
        self.furthest_bar = int(max(current_end_times) / 4) if current_end_times else 0
        return self.furthest_bar

    @above_version(supported_version=(8, 2, 0))
    def get_bpm(self) -> float:
        """Get bpm from Ableton Live set XML."""
        if self.version_tuple is None:
            raise SetError("Set version is not parsed!")
        post_10_bpm = "LiveSet.MasterTrack.DeviceChain.Mixer.Tempo.Manual"
        pre_10_bpm = "LiveSet.MasterTrack.MasterChain.Mixer.Tempo.ArrangerAutomation.Events.FloatEvent"
        pre_10_bpm = "LiveSet.MasterTrack.DeviceChain.Mixer.Tempo.ArrangerAutomation.Events.FloatEvent"
        major, minor, _ = self.version_tuple
        if major >= 10 or major >= 9 and minor >= 7:
            bpm_elem = get_element(self.root, post_10_bpm, attribute="Value", silent_error=True)
        else:
            bpm_elem = get_element(self.root, pre_10_bpm, attribute="Value")
        self.bpm = round(float(bpm_elem), 6)
        return self.bpm

    def estimate_length(self) -> None:
        """Multiply the longest bar with length per bar by inverting BPM."""
        if self.bpm is None or self.furthest_bar is None:
            logger.info("%sCan't estimate length without bpm and furthest bar.", R)
            return
        # TODO improve this to find the time signature from the set and use it here instead of only 4/4.
        seconds_total = ((4 * int(self.furthest_bar)) / self.bpm) * 60
        length = f"{int(seconds_total // 60)}:{round(seconds_total % 60):02d}"
        logger.info(
            "%sLongest clip or furthest arrangement position: %s bars. %sEstimated length(Only valid for 4/4): %s",
            M,
            self.furthest_bar,
            C,
            length,
        )

    @set_loaded
    def set_track_heights(self, height: int) -> None:
        """In Arrangement view, sets all track lanes/automation lanes to specified height."""
        assert self.root is not None  # Shutup mypy, not possible at runtime.
        height = min(425, (max(17, height)))  # Clamp to valid range.
        for el in self.root.iter("LaneHeight"):
            el.set("Value", str(height))
        logger.info("%sSet track heights to %s.", G, height)

    @set_loaded
    def set_track_widths(self, width: int) -> None:
        """Set all track widths in Clip view to specified width."""
        assert self.root is not None  # Shutup mypy.
        width = min(264, (max(17, width)))  # Clamp to valid range.
        # Sesstion is how it's named in the set, not a typo!
        for el in self.root.iter("ViewStateSesstionTrackWidth"):
            el.set("Value", str(width))
        logger.info("%sSet track widths to %s.", G, width)

    @set_loaded
    def fold_tracks(self) -> None:
        """Fold all tracks."""
        assert self.root is not None  # Shutup mypy.
        for el in self.root.iter("TrackUnfolded"):
            el.set("Value", "false")
        logger.info("%sFolded all tracks.", G)

    @set_loaded
    def unfold_tracks(self) -> None:
        """Unfold all tracks."""
        assert self.root is not None  # Shutup mypy, not possible at runtime.
        for el in self.root.iter("TrackUnfolded"):
            el.set("Value", "true")
        logger.info("%sUnfolded all tracks.", G)

    @above_version(supported_version=(8, 2, 0))
    def set_audio_output(self, output_number: int, element_string: str) -> None:
        """Set audio output."""
        if output_number not in STEREO_OUTPUTS:
            raise ValueError(f"{R}Output number invalid!. Available options: \n{STEREO_OUTPUTS}{RST}")
        output_obj = STEREO_OUTPUTS[output_number]
        out_target_element = get_element(
            self.root,
            f"LiveSet.{element_string}.DeviceChain.AudioOutputRouting.Target",
            silent_error=True,
        )
        if not isinstance(out_target_element, ET.Element):
            out_target_element = get_element(  # ableton 8 sets use "MasterChain" for master track.
                self.root,
                f"LiveSet.{element_string}.MasterChain.AudioOutputRouting.Target",
            )
            lower_display_string_element = get_element(
                self.root,
                f"LiveSet.{element_string}.MasterChain.AudioOutputRouting.LowerDisplayString",
            )
        else:
            lower_display_string_element = get_element(
                self.root,
                f"LiveSet.{element_string}.DeviceChain.AudioOutputRouting.LowerDisplayString",
            )
        out_target_element.set("Value", output_obj["target"])
        lower_display_string_element.set("Value", output_obj["lower_display_string"])
        logger.info("%sSet %s to %s", G, element_string, output_obj["lower_display_string"])

    def _parse_hex_path(self, text: str) -> Optional[str]:
        """Take raw hex string from XML entry and parses."""
        if not text:
            return None
        # Strip new lines and tabs from raw text to have one long hex string.
        abs_hash_path = text.replace("\t", "").replace("\n", "")
        byte_data = bytearray.fromhex(abs_hash_path)
        if byte_data[0:3] == b"\x00" * 3:  # Header only on mac projects.
            self.set_os = SetOperatingSystem.MAC_OS
            return utils.parse_mac_data(byte_data, abs_hash_path)
        else:
            self.set_os = SetOperatingSystem.WINDOWS_OS
            return utils.parse_windows_data(byte_data, abs_hash_path)

    def path_separator_type(self, path_str: str) -> str:
        """Get OS path string separator."""
        # TODO: Move this into utils.
        if "\\" in path_str:
            return "\\"
        elif "/" in path_str:
            return "/"
        else:
            raise ValueError(f"Couldn't parse OS path type! {path_str}")

    def search_plugins(self, plugin_name: str) -> Optional[pathlib.Path]:
        """Search for plugins and add them to self.found_vst_dirs."""
        if sys.platform == "win32":
            drive = os.environ["SYSTEMDRIVE"]
            _WINDOWS_VST3 = pathlib.Path(rf"{drive}\Program Files\Common Files\VST3")
            vst3_plugins = list(_WINDOWS_VST3.rglob("*.dll")) + list(_WINDOWS_VST3.rglob("*.vst3"))
            for vst3 in vst3_plugins:
                if plugin_name == vst3.name:
                    return vst3
            for directory in self.found_vst_dirs:
                for dll in directory.rglob("*.dll"):
                    if plugin_name == dll.name or plugin_name == dll.name.replace(".32", "").replace(".64", ""):
                        return dll
            return None
        else:
            # TODO: Implement MacOS VST3 logic
            logger.warning("%sMac OS Vst3 not implemented yet.", RB)
            return None

    # Plugin related functions.
    def parse_vst_element(
        self, vst_element: ET.Element
    ) -> Tuple[Optional[pathlib.Path], Optional[str], Optional[pathlib.Path]]:
        """Parse out VST element from vst xtree."""
        for plugin_path in ["Dir", "Path"]:
            path_results = vst_element.findall(f".//{plugin_path}")
            if len(path_results):
                if plugin_path == "Path":
                    if (full_path := path_results[0].get("Value")) is None:
                        logger.error("Couldn't get Path for %s", path_results[0])
                        continue
                    if not "/" in full_path and not "\\" in full_path:
                        if search_result := self.search_plugins(full_path):
                            return None, search_result.name, search_result
                        return None, full_path, None
                    path_separator = self.path_separator_type(full_path)
                    name = full_path.split(path_separator)[-1]
                    return pathlib.Path(full_path), name, None
                elif plugin_path == "Dir":
                    if (dir_bin := path_results[0].find("Data")) is None:
                        logger.error("Couldn't get Path for %s", path_results[0])
                        continue
                    if (text := dir_bin.text) is None:
                        continue
                    path = self._parse_hex_path(text)
                    name_ele = vst_element.find("FileName")
                    name = name_ele.get("Value", "") if name_ele is not None else "<>"
                    if not path:
                        logger.error("%sCouldn't parse absolute path for %s", Y, name)
                        return None, name, None
                    path_separator = self.path_separator_type(path)
                    if path[-1] == path_separator:
                        full_path = f"{path}{name}"
                    else:
                        full_path = f"{path}{path_separator}{name}"
                    return pathlib.Path(full_path), name, None

        logger.error("%sCouldn't parse plugin!", R)
        return None, None, None

    def list_plugins(self, vst_dirs: List[pathlib.Path]) -> List[pathlib.Path]:
        """Iterates through all plugin references and checks paths for VSTs."""
        self.found_vst_dirs.extend(vst_dirs)
        for plugin_element in self.root.iter("PluginDesc"):
            self.last_elem = plugin_element
            for vst_element in plugin_element.iter("VstPluginInfo"):
                full_path, name, potential = self.parse_vst_element(vst_element)
                exists = True if full_path and full_path.exists() else False
                if exists and full_path.parent not in self.found_vst_dirs:
                    self.found_vst_dirs.append(full_path.parent)
                elif not exists:
                    # Did not find plugin in saved path, try to search
                    potential = self.search_plugins(name)
                color = G if exists else R
                if potential and color == R:
                    color = Y
                logger.info(
                    "%sPlugin: %s, %sPlugin folder path: %s, %sExists: %s", color, name, M, full_path, color, exists
                )
                if potential:
                    logger.info("%s\tPotential alternative path for %s found: %s%s", CB, name, M, potential)
            for au_element in plugin_element.iter("AuPluginInfo"):
                name = au_element.find("Name").get("Value")
                manufacturer = get_element(plugin_element, "AuPluginInfo.Manufacturer", attribute="Value")
                logger.info(
                    "%sMac OS Audio Units are not saved with paths. Plugin %s: %s cannot be verified.",
                    M,
                    manufacturer,
                    name,
                )
                # TODO figure out how to match different name from components installed to stored set plugin.
                # au_components = pathlib.Path('/Library/Audio/Plug-Ins/Components').rglob('*.component')
        return self.found_vst_dirs

    def _list_samples(self) -> None:
        """Post Ableton 11 sample parser. Format changed from binary encoded paths to simple strings for all OSes."""
        missing_samples = 0
        for parsed in self._iterate_samples():
            if parsed.absolute_exists or parsed.relative_exists:
                # Sample will load in ableton, no need to do anything.
                logger.debug(
                    "%sSample %s found: Relative %s, Absolute %s",
                    G,
                    parsed.name,
                    parsed.relative,
                    parsed.absolute,
                )
                continue
            missing_samples += 1
            logger.warning(
                "%sSample %s missing: \n\tAbsolute[%s], Relative [%s]", R, parsed.name, parsed.absolute, parsed.relative
            )

    def list_samples(self) -> None:
        """Select correct sample parsing function."""
        if self.project_root_folder is None:
            self.find_project_root_folder()
        if self.version_tuple is None:
            self.load_version()
        if self.version_tuple is None:
            raise SetError("Version not parsed!")
        return self._list_samples()

    def _parse_samplepaths(
        self,
        absolute_path: Optional[pathlib.Path],
        relative_path: Optional[pathlib.Path],
        verbose: bool,
        saved_filesize: int,
    ) -> bool:
        absolute_found = absolute_path is not None and absolute_path.exists()
        relative_found = relative_path is not None and relative_path.exists()
        if not absolute_found and not relative_found:
            if absolute_path and absolute_path not in self.missing_absolute_samples:
                self.missing_absolute_samples.append(absolute_path)
            if relative_path and relative_path not in self.missing_relative_samples:
                self.missing_relative_samples.append(relative_path)
            return False
        if absolute_found and absolute_path is not None:
            local_filesize = absolute_path.stat().st_size
            if verbose:
                size_match = saved_filesize == local_filesize
                logger.info(
                    "%sAbsolute path sample found: %s\n\tFile size %s matches saved filesize %s: %s%s",
                    G,
                    absolute_path,
                    local_filesize,
                    saved_filesize,
                    G if size_match else R,
                    size_match,
                )
        if relative_found and relative_path is not None:
            local_filesize = relative_path.stat().st_size
            size_match = saved_filesize == local_filesize
            if verbose:
                logger.info(
                    "%sRelative(collect and save) sample found: %s\n\tFile size %s matches saved filesize %s: %s%s",
                    G,
                    relative_path,
                    local_filesize,
                    saved_filesize,
                    G if size_match else R,
                    size_match,
                )
        if absolute_found or relative_found:
            return True
        return False

    def sample_results(self, missing_samples: int) -> None:
        """logger.info results of sample search."""
        color = G if not missing_samples else Y
        logger.info(
            "%sTotal missing sample references: %s%s%s, this can include duplicate references to the same sample so "
            "only unique paths are listed here. Relative paths are created using collect-and-save. If either sample "
            "path is found Ableton will load the sample.",
            color,
            M,
            missing_samples,
            color,
        )
        if self.missing_relative_samples:
            rel_string = "\n\t".join((str(x) for x in self.missing_relative_samples))
            logger.info("%sMissing Relative paths:%s\n\t%s", Y, R, rel_string)
        if self.missing_absolute_samples:
            abs_string = "\n\t".join((str(x) for x in self.missing_absolute_samples))
            logger.info("%sMissing Absolute paths:%s\n\t%s", Y, R, abs_string)

    def _iterate_samples(self) -> Generator[utils.SampleRef, None, None]:
        """Iterate through set sample references and build sample list."""
        if self.sample_list:
            for parsed in self.sample_list:
                yield parsed
            return
        for sample_ref in self.root.iter("SampleRef"):
            parsed = utils.SampleRef.from_element(sample_ref, self.version_tuple, self.project_root_folder)
            self.sample_list.append(parsed)
            yield parsed
        return

    def fix_samples(self, db: Dict[str, Dict[str, str]], collect_and_save: bool = False, force: bool = False) -> bool:
        """Fix broken sample paths.

        Args:
            db: database loaded from json.
            collect_and_save: copy any found samples into the project folder, the same as ableton's collect
                and save
            force: used with collect_and_save. When the same name sample is found in the project, force replace
                it if the project's current file is a different file size.
        """
        self.find_project_root_folder()
        found_samples: Dict[str, Dict[str, str]] = {}
        missing_samples = 0
        fixed_samples = 0
        skip_search = False
        for parsed in self._iterate_samples():
            if parsed.absolute_exists or parsed.relative_exists:
                # Sample will load in ableton, no need to do anything.
                continue
            missing_samples += 1

            # Skip builtin pack content for now. Can revisit this later but these samples probably will fix
            # automatically in ableton on set load.
            factory_packs = ["/Resources/Builtin/Samples", "Ableton/Factory Packs"]
            if any([x in str(parsed.absolute.parent) for x in factory_packs]):
                logger.debug("%sSkipping builtin pack content: %s", Y, parsed.absolute)
                continue

            # There's often the same sample referenced many times in the same set, check previous found first.
            for smp_path, smp_info in found_samples.items():
                if self._fix_sample(collect_and_save, parsed, smp_info, smp_path, found_samples, force):
                    fixed_samples += 1
                    skip_search = True
                    break
            if skip_search:
                skip_search = False
                continue
            # Iterating through hashes is extremely fast :D
            for smp_path, smp_info in db.items():
                if self._fix_sample(collect_and_save, parsed, smp_info, smp_path, found_samples, force):
                    fixed_samples += 1
                    break
            else:
                logger.warning(
                    "%sCould not find sample for %s\n%s\n%s", Y, parsed.name, parsed.absolute, parsed.relative
                )

        logger.info(
            "%sOrignal missing sample count: %s, Samples fixed: %s, Couldn't fix: %s",
            G if fixed_samples == missing_samples else R,
            missing_samples,
            fixed_samples,
            missing_samples - fixed_samples,
        )

    def _fix_sample(
        self,
        collect_and_save: bool,
        parsed: utils.SampleRef,
        smp_info: Dict[str, str],
        smp_path: str,
        found_samples: Dict[str, Dict[str, str]],
        force: bool,
    ) -> bool:
        """Attempt to fix sample if matches DB entry.

        size is not always stored in ableton sets unfortunately, but we do usually have last_modified.
        This is not perfect, but the probability of a file name matching and it's last modification time
        matching and being a false positive are quite low.
        """
        if smp_info.get("name") != parsed.name:
            return False
        size_match = parsed.size and smp_info.get("size") == parsed.size
        modified_match = parsed.last_modified and parsed.last_modified == int(smp_info.get("last_modified"))
        if not size_match and not modified_match:
            return False

        logger.debug("\n\n%sFound potential match %s, \n[%s]\n%s%s", G, smp_path, smp_info, M, parsed)
        found_samples[smp_path] = smp_info
        replacement_sample = pathlib.Path(smp_path)

        if collect_and_save and self.project_root_folder:
            # Relative type 3 is collected and saved, 1 is absolute path.
            relative_type = parsed.get_relative_type()
            if relative_type == 3:
                rel_path = str(parsed.get_relative_value())
            else:
                rel_path = "Samples/Imported"
            (self.project_root_folder / rel_path).mkdir(parents=True, exist_ok=True)

            copied_sample = self.project_root_folder / rel_path / smp_info.get("name")
            if copied_sample.exists() and copied_sample.stat().st_size != parsed.size:
                logger.error(
                    "%sCannot copy sample %s, would replace existing one in project with " "same name! Skipping...",
                    R,
                    copied_sample,
                )
                return False
            elif copied_sample.exists() and copied_sample.stat().st_size == parsed.size:
                pass
            else:
                shutil.copy(replacement_sample, copied_sample)
            parsed.set_relative(f"{rel_path}/{copied_sample.name}")
            parsed.set_relative_type(3)
        elif collect_and_save and not self.project_root_folder:
            logger.warning(
                "%sProject root () not found, can't collect and save this sample. " "Using absolute path instead..", Y
            )
            parsed.set_absolute(replacement_sample)
            parsed.set_relative_type(1)
        else:
            parsed.set_absolute(replacement_sample)
            parsed.set_relative_type(1)
        return True

    def gradient_tracks(self) -> None:
        """Make a rough gradient across tracks using built in colors."""
        if not self.tracks:
            self.load_tracks()
        for clr_ind, track in zip(color_tools.create_gradient_ableton(len(self.tracks)), self.tracks):
            track.color = clr_ind

            clipview_clr_elements = list(track.clip_clipview_colors())
            clip_view_gradient = color_tools.create_gradient_ableton(len(clipview_clr_elements), starting_index=clr_ind)
            for sub_ind, clip_clr_ele in zip(clip_view_gradient, clipview_clr_elements):
                clip_clr_ele.set("Value", str(sub_ind))

            arangement_clr_elements = list(track.clip_arangement_colors())
            clip_view_gradient = color_tools.create_gradient_ableton(
                len(arangement_clr_elements), starting_index=clr_ind
            )
            for sub_ind, clip_clr_ele in zip(clip_view_gradient, arangement_clr_elements):
                clip_clr_ele.set("Value", str(sub_ind))

    @above_version(supported_version=(9, 0, 0))
    def plugin_buffers(self, verbose: bool) -> None:
        for plugin in self.root.iter("VstPluginInfo"):
            for path_query in ["Path", "Dir"]:
                path_ele = plugin.findall(f".//{path_query}")
                if len(path_ele):
                    if path_query == "Dir":
                        data = path_ele[0].findall(".//Data")
                        path_str = self._parse_hex_path(data[0].text)
                        break
                    else:
                        path_str = path_ele[0].get("Value")

            name = plugin.find("PlugName").get("Value")
            buffer_str = plugin.findall(".//Buffer")[0].text
            if not buffer_str:
                continue

            parsed_buf = buffer_str.replace("\t", "").replace("\n", "")
            buffer_bytes = bytes.fromhex(parsed_buf)
            with open(self.pathlib_obj.parent / f"{name}_{self.name}.plugin_dump", "w") as f:
                f.write("ASCII:\n\n")
                f.write(buffer_bytes.decode("ascii", errors="ignore"))
                f.write("\n\nHEX:\n\n")
                f.write(parsed_buf)

            encodings = ["ascii", "utf-8", "utf-16"]
            detected_enc = chardet.detect(buffer_bytes)["encoding"]
            if detected_enc:
                encodings.insert(0, detected_enc)

            for encoding in encodings:
                try:
                    decoded = buffer_bytes.decode(encoding=encoding, errors="ignore")
                    logging.info(
                        name,
                        encoding,
                        decoded[:500] if len(decoded) >= 1000 else decoded,
                    )
                except UnicodeDecodeError as exc:
                    if verbose:
                        logging.info(f"{R}Couldn't decode with {encoding} for bytes: {exc}, data:\n{buffer_bytes}")
                    decoded = "COULD NOT DECODE"

            end_index = min(len(decoded), 800) if not verbose else len(decoded)
            if not verbose and len(decoded) > 800:
                logging.info(
                    f"{G}Name: {name}, {B}Path: {path_str}, {C}Encoding: {detected_enc}, "
                    f"Buffer(truncated if more than 800 characters):\n{decoded[:end_index]}"
                )
            else:
                logging.info(
                    f"{G}Name: {name}, {B}Path: {path_str}, {C}Encoding: {detected_enc}, Buffer:\n{decoded}\n\n"
                )



"""Cli entry point."""

import argparse
import copy
import datetime
import logging
import os
import pathlib
import sys
import time
import traceback
from typing import Dict, List, Optional

from abletoolz import __version__
from abletoolz.ableton_set import AbletonSet
from abletoolz.misc import BACKUP_DIR, CB, B, C, ElementNotFound, M, R, Y
from abletoolz.sample_databaser import create_db

logger = logging.getLogger(__name__)


def get_pathlib_objects(srcs: List[str]) -> List[pathlib.Path]:
    """Get all ableton sets to parse.

    Args:
        srcs: path or paths to directories and set files.

    Returns:
        list of pathlib.Paths with all ableton sets to parse, excluding backup directories.
    """
    paths: List[pathlib.Path] = []
    for src in srcs:
        path = pathlib.Path(src)
        if path.is_dir():
            files = list(path.rglob("*.als"))
            # Hacky but Path.rglob doesn't have options for filtering.
            files_to_process = []
            for file in files:
                if all(x not in file.parts[:-1] for x in ["Backup", "backup", BACKUP_DIR]) and not file.stem.startswith(
                    ("._")
                ):
                    files_to_process.append(file)
            paths.extend(files_to_process)
        elif path.is_file():
            paths.append(path)
    return paths


def is_valid_dir_path(path: str) -> str:
    """Check if the path is a valid.

    Mainly for windows, which uses backslashes instead and this causes problems for parsing command line arguments since
    backslash is used for escaping.
    """
    if sys.platform.startswith("win") and '"' in path:
        raise ValueError(
            f"{R}Windows paths must not end in backslash: \n'C:\\somepath\\'(BAD)\n'C:\\somepath' "
            f"(GOOD)\nThis is due to a bug in how Windows handles backslashes before quotes."
        )
    return path


def parse_arguments() -> argparse.Namespace:
    """Get command line arguments."""
    parser = argparse.ArgumentParser(description=f"abletoolz {__version__}", add_help=True)

    # Input arguments.
    parser.add_argument(
        "srcs",
        action="store",
        nargs="*",
        help="Set(s) or directory(ies). All sub folders in directories are parsed for sets. NOTE: On WINDOWS remove "
        "the trailing backslash when processing folders! This is due to how windows and python interact with "
        "backslashes, which are normally escape characters.",
    )
    parser.add_argument(
        "--db",
        "--database",
        action="store_true",
        default=False,
        help="Instead of parsing sets, create/update sample database for fast lookups when fixing broken paths.",
    )

    # Output arguments.
    parser.add_argument(
        "-x",
        "--xml",
        action="store_true",
        default=False,
        help="dump the xml in same directory as set_name.xml(Useful to understand set structure).",
    )
    parser.add_argument(
        "-s",
        "--save",
        action="store_true",
        default=False,
        help="Saves file after parsing. This is only put here as a safety check, to make sure you know "
        "what you are doing! The original file is always renamed to "
        f"set_directory/{BACKUP_DIR}/set_name_xx.als, where xx will automatically increase to "
        "to avoid overwriting files.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="Adds extra verbosity, for some commands this will print more information.",
    )

    parser.add_argument(
        "--append-bars-bpm",
        action="store_true",
        default=False,
        help="Append furthest bar length and bpm to filename to help organize your set collection. "
        "For example, my_set.als --> my_set_32bars_90.00bpm.als Option only works with -s/--save",
    )
    parser.add_argument(
        "--prepend-version",
        action="store_true",
        default=False,
        help="Appends set version to set filename",
    )

    # Edit set arguments.
    parser.add_argument(
        "--unfold",
        action="store_true",
        default=False,
        help="unfolds all tracks/automation lanes in arrangement.",
    )
    parser.add_argument(
        "--fold",
        action="store_true",
        default=False,
        help="folds all tracks/automation lanes in arrangement.",
    )
    parser.add_argument("--set-track-heights", type=int, help="Set arrangement track heights")
    parser.add_argument("--set-track-widths", type=int, help="Set clip view track width.")
    parser.add_argument(
        "--cue-out",
        type=int,
        help="set Cue audio output tracks. Set to 1 for stereo 1/2, 2 for 3/4 etc",
    )
    parser.add_argument(
        "--master-out",
        type=int,
        help="number to set Master audio output tracks to. Same numbers as --cue-out",
    )
    parser.add_argument(
        "--fix-samples-absolute",
        action="store_true",
        default=False,
        help="Finds missing samples and fixes the broken references in your ableton sets. Does not copy sample into "
        "project folder. Run --db on folders first "
        "to create your database.",
    )
    parser.add_argument(
        "--fix-samples-collect",
        action="store_true",
        default=False,
        help="Collects and saves missing samples into the set's project folder, the same as collect and save in "
        "Ableton. Run --db on folders first to create your database.",
    )
    parser.add_argument(
        "--gradient-tracks",
        action="store_true",
        default=False,
        help="Generate a random gradient over the tracks and color them. Ableton has a very limited set of available "
        "colors, so the results are limited, but you still can get some decent results. This uses the CIE2000 "
        "algorithm which helps create a gradient more natural to the human eye.",
    )

    # Analysis arguments.
    parser.add_argument(
        "--list-tracks",
        action="store_true",
        default=False,
        help="Load and list all track information.",
    )
    parser.add_argument(
        "--check-samples",
        action="store_true",
        default=False,
        help="Checks relative and absolute sample paths and verifies if sample exists there.",
    )

    parser.add_argument(
        "--check-plugins",
        action="store_true",
        default=False,
        help="Checks plugin VST paths and verifies they exists. Note: If Ableton finds the "
        "plugin name in a different path it will automatically update these paths the next time "
        "you save your project, so take it with a grain of salt. AU are not stored as paths in "
        "sets but abbreviated component names. Might possibly add support for them later.",
    )
    parser.add_argument(
        "--plugin-buffers",
        action="store_true",
        default=False,
        help="Dump raw hex binary data from plugin buffers.",
    )

    args = parser.parse_args()
    assert not (
        args.fix_samples_absolute and args.fix_samples_collect
    ), "Can only use --fix-samples-collect or --fix-samples-absolute, not both!"
    assert not (args.fold and args.unfold), "Only set unfold or fold, not both."
    assert not (
        args.db
        and any(
            [
                args.append_bars_bpm,
                args.check_plugins,
                args.check_samples,
                args.cue_out,
                args.fix_samples_absolute,
                args.fix_samples_collect,
                args.fold,
                args.gradient_tracks,
                args.list_tracks,
                args.master_out,
                args.prepend_version,
                args.plugin_buffers,
                args.save,
                args.set_track_heights,
                args.set_track_widths,
                args.unfold,
                args.xml,
            ]
        )
    ), "--db/--database cannot be used with other commands!"
    return args


def process_set(args: argparse.Namespace, pathlib_obj: pathlib.Path, db: Optional[Dict]) -> int:
    """Process individual set."""
    logger.info("%sParsing: %s", C, pathlib_obj)
    ableton_set = AbletonSet(pathlib_obj)
    if not ableton_set.parse():
        return -2
    ableton_set.load_version()
    logger.info("%sSet name: %s, %sBPM: %s", C, pathlib_obj.stem, B, ableton_set.get_bpm())
    ableton_set.find_project_root_folder()
    ableton_set.find_furthest_bar()
    ableton_set.estimate_length()
    # if args.save:
    #     backup = copy.deepcopy(ableton_set)

    if args.master_out:
        ableton_set.set_audio_output(args.master_out, element_string="MasterTrack")
    if args.cue_out:
        ableton_set.set_audio_output(args.cue_out, element_string="PreHearTrack")
    if args.fold:
        ableton_set.fold_tracks()
    elif args.unfold:
        ableton_set.unfold_tracks()
    if args.set_track_heights:
        ableton_set.set_track_heights(args.set_track_heights)
    if args.set_track_widths:
        ableton_set.set_track_widths(args.set_track_widths)
    if args.gradient_tracks:
        ableton_set.gradient_tracks()

    if args.list_tracks:
        ableton_set.load_tracks()
        ableton_set.print_tracks()

    if args.check_samples:
        ableton_set.list_samples()

    if args.fix_samples_absolute or args.fix_samples_collect:
        ableton_set.fix_samples(db, collect_and_save=args.fix_samples_collect)

    if args.check_plugins:
        vst_dirs: List[pathlib.Path] = []
        result = ableton_set.list_plugins(vst_dirs)
        vst_dirs = list(set(result))
    if args.plugin_buffers:
        ableton_set.plugin_buffers()


    if args.xml:
        ableton_set.save_xml()
    if args.save:
        # if backup == ableton_set:
        #     logger.info("%sSet has no changes from originally, not saving...", MB)
        #     return 0
        ableton_set.save_set(append_bars_bpm=args.append_bars_bpm, prepend_version=args.prepend_version)
    elif any(
        [
            args.append_bars_bpm,
            args.check_plugins,
            args.check_samples,
            args.cue_out,
            args.fix_samples_absolute,
            args.fix_samples_collect,
            args.fold,
            args.gradient_tracks,
            args.master_out,
            args.prepend_version,
            args.set_track_heights,
            args.set_track_widths,
            args.unfold,
        ]
    ):
        logger.info("%sNo changes saved, use -s/--save option to write changes to file.", Y)
    return 0


def process(args: argparse.Namespace) -> int:
    """Process arguments.

    Args:
        args: argparse.Namespace with parsed arguments.

    Returns:
        integer with exit code, zero indicating success, non-zero indicating error.
    """
    if args.db:
        create_db.create_or_update_db(args.srcs)
        return 0
    db = None
    if args.fix_samples_collect or args.fix_samples_absolute:
        logger.info("%sLoading db...", M)
        db = create_db.load_db()

    start_time = time.time()
    pathlib_objects = get_pathlib_objects(srcs=args.srcs)
    if not pathlib_objects:
        logger.info("%sError, no sets to process!", R)
        return -1

    for pathlib_obj in pathlib_objects:
        try:
            process_set(args, pathlib_obj, db)
        except ElementNotFound:
            logger.info(traceback.format_exc())
        logger.info("%s\n\n%s\n\n", M, "^" * os.get_terminal_size().columns)
    logger.info(
        "%sTook %s to process %s set(s)", CB, datetime.timedelta(seconds=time.time() - start_time), len(pathlib_objects)
    )
    return 0


def main() -> None:
    """Entry point to cli."""
    args = parse_arguments()

    logging.getLogger("colormath").setLevel(logging.WARNING)
    logging.basicConfig(
        stream=sys.stdout,
        level=logging.DEBUG if args.verbose else logging.INFO,
        format="%(message)s",
        datefmt="%H:%M:%S",
    )
    sys.exit(process(args))


if __name__ == "__main__":
    main()



"""Create sample database for fast searches and fixing broken sample paths."""

import json
import logging
import pathlib
from typing import Dict, List, Optional

import tqdm

from abletoolz.misc import DEFAULT_DB_PATH

logger = logging.getLogger(__name__)


def get_all_audio_files(path: pathlib.Path) -> List[pathlib.Path]:
    """Find all supported audio files in directory."""
    file_suffixes = [
        "*.aiff",
        "*.AIFF",
        "*.aif",
        "*.AIF",
        "*.wav",
        "*.WAV",
        "*.mp3",
        "*.MP3",
        "*.flacc",
        "*.FLACC",
        "*.ogg",
        "*.OGG",
        "*.mp4",
        "*.MP4",
    ]
    all_files: List[pathlib.Path] = []
    for file_type in file_suffixes:
        all_files.extend(list(path.rglob(file_type)))
    return all_files


def create_or_update_db(paths: List[str], db_path: Optional[pathlib.Path] = None) -> pathlib.Path:
    """Search all samples and add to database, create new one if it doesn't exist.

    db {
        path_string: {
            file_size: sample file size,
            # Crc unsupported for now, can't figure out ableton's crc algorithm! This would allow for
            # near perfect matches, although file size and perfectly matching file name is probably good enough.
            crc: crc size
            last_modified: time of last file modification.
        }
    }
    """
    if db_path is None:
        db_path = DEFAULT_DB_PATH
    logger.info(
        "Using db path %s. Creating database from scratch can take a while, please be patient. Updating "
        "an existing one is much faster!",
        db_path.resolve(),
    )
    if not db_path.exists():
        db = {}
    else:
        with db_path.open("r") as f:
            db = json.load(f)
    all_files = []
    for path in paths:
        all_files.extend(get_all_audio_files(pathlib.Path(path)))

    # Remove any broken paths.
    to_remove: List[str] = []
    for path in tqdm.tqdm(db.keys(), desc="Validating current db..."):
        if not pathlib.Path(path).exists():
            to_remove.append(path)
    for path in to_remove:
        db.pop(path)

    for sample in tqdm.tqdm(all_files, desc="Progress"):
        if str(sample.resolve()) in db:
            continue
        db[str(sample.resolve())] = {
            "name": sample.name,
            "size": sample.stat().st_size,
            "last_modified": sample.stat().st_mtime,
        }

    # Write db!
    with db_path.open("w") as f:
        json.dump(db, f, indent=2, sort_keys=True)
    logger.info("Updated database at %s", db_path.resolve())
    return db_path


def load_db() -> Dict[str, Dict[str, Dict[str, str]]]:
    """Load db from json."""
    if not DEFAULT_DB_PATH.exists():
        raise FileNotFoundError(f"Database {DEFAULT_DB_PATH} doesn't exist! Run --db with sample dir(s) first.")
    with DEFAULT_DB_PATH.open() as f:
        return json.load(f)
